#include <string>
#include <vector>
#include <cassert>

#include "OsFile.h"

#include <curl/curl.h>

#define	DEFAULT_TIMEOUT_MILLISECS	100
#define	MAX_TIMEOUT_MILLISECS		500

static void CurlShareLockCallback( CURL* handle, curl_lock_data data, curl_lock_access access, void* useptr )
{
	::WaitForSingleObject( static_cast<HANDLE>( useptr ), INFINITE );
}

static void CurlShareUnlockCallback( CURL* handle, curl_lock_data data, void* useptr )
{
	::ReleaseMutex( static_cast<HANDLE>( useptr ) );
}

static size_t CurlReadCallback( void *ptr, size_t size, size_t nmemb, void *stream )
{
	OsFile* file = static_cast<OsFile*>( stream );
	assert( file != 0 );

	// is this file open allready?
	if( !file->IsOpen() ) 
	{
		if( !file->Open( "rb" ) )
			return( CURL_READFUNC_ABORT );
	}

	// try to read as many data as fits into the buffer:
	size_t bytesToRead = size * nmemb;
	size_t bytesRead = fread( ptr, 1, bytesToRead, *file );
	if( bytesRead == 0 )
	{
		file->Close();
	}

	return( bytesRead );
}

static bool Upload( int numChannels, bool activeFtp, OsFileVector& files, const std::string& host, const std::string& path, const std::string& user, const std::string& password )
{
	assert( numChannels >= 1 );
	assert( numChannels <= 4 );

	const std::string userPassword = user + ":" + password;

	// create share handle:
	//
	HANDLE g_hMutex = CreateMutex( NULL, FALSE, NULL );
	assert( g_hMutex );
	CURLSH* curlShare = curl_share_init();
	assert( curlShare != 0 );
	curl_share_setopt( curlShare, CURLSHOPT_LOCKFUNC,   CurlShareLockCallback );
	curl_share_setopt( curlShare, CURLSHOPT_UNLOCKFUNC, CurlShareUnlockCallback );
	curl_share_setopt( curlShare, CURLSHOPT_USERDATA,   g_hMutex );
	curl_share_setopt( curlShare, CURLSHOPT_SHARE,      CURL_LOCK_DATA_DNS );

	if( files.size() < (OsFileVector::size_type)numChannels )
		numChannels = (int)files.size();

	// create multi handle:
	//
	CURLM* curlMulti = curl_multi_init();
	assert( curlMulti != 0 );
	curl_multi_setopt( curlMulti, CURLMOPT_MAXCONNECTS, numChannels );

	// prepare libcurl easy handles and add to multi handle:
	//
	CURL** curlEasy = new CURL*[ numChannels ];
	assert( curlEasy != 0 );

	OsFileVector::const_iterator file = files.begin();
	for( int i = 0; i < numChannels; i++ )
	{
		CURL* easy = curl_easy_init();
		assert( easy != 0 );

		std::string url = host + path + files.GetRelativePath( (*file)->GetPathname() ).c_str();

		curl_easy_setopt( easy, CURLOPT_UPLOAD                   , 1                                           );
		curl_easy_setopt( easy, CURLOPT_SHARE                    , curlShare                                   );
		curl_easy_setopt( easy, CURLOPT_NOSIGNAL                 , 1                                           );
		curl_easy_setopt( easy, CURLOPT_NOPROGRESS               , 1  /* set to 1 to disable progress meter */ );
		curl_easy_setopt( easy, CURLOPT_VERBOSE                  , 1                                           );
		curl_easy_setopt( easy, CURLOPT_DNS_USE_GLOBAL_CACHE     , 0                                           );
		curl_easy_setopt( easy, CURLOPT_FTP_USE_EPRT             , 0  /* use PORT instead */                   );
		curl_easy_setopt( easy, CURLOPT_FTP_USE_EPSV             , 0  /* use PASV instead */                   );
		curl_easy_setopt( easy, CURLOPT_FTP_FILEMETHOD           , CURLFTPMETHOD_MULTICWD                      );
		curl_easy_setopt( easy, CURLOPT_FTP_CREATE_MISSING_DIRS  , 1                                           );
		curl_easy_setopt( easy, CURLOPT_SSH_AUTH_TYPES           , CURLSSH_AUTH_PASSWORD                       );
		curl_easy_setopt( easy, CURLOPT_USERPWD                  , userPassword.c_str()                        );
		curl_easy_setopt( easy, CURLOPT_FTPPORT                  , activeFtp ? "-" : 0                         );

		__int64 fileSize;
		bool fileSizeAcquired = (*file)->GetSize( fileSize );
		if( fileSizeAcquired )
			curl_easy_setopt( easy, CURLOPT_INFILESIZE_LARGE     , fileSize                                    );
		curl_easy_setopt( easy, CURLOPT_URL                      , url.c_str()                                 );
		curl_easy_setopt( easy, CURLOPT_READFUNCTION             , CurlReadCallback                            );
		curl_easy_setopt( easy, CURLOPT_READDATA                 , *file                                       );
		curl_multi_add_handle( curlMulti, easy );
		file++;

		curlEasy[ i ] = easy;
	}

	int running_handles;
	CURLMcode multiResult;
	bool result = true;

	while( result )
	{

		// call curl_multi_perform, as long as it wants to be called:
		// 
		while( ( multiResult = curl_multi_perform( curlMulti, &running_handles ) ) == CURLM_CALL_MULTI_PERFORM ) ;


		// now process waiting easy handle messages:
		//
		int num_messages;
		CURLMsg* curlMsg = curl_multi_info_read( curlMulti, &num_messages );
		while( curlMsg ) {
			if( curlMsg->msg == CURLMSG_DONE )
			{
				CURL* easy = curlMsg->easy_handle;
				CURLcode easyResult = curlMsg->data.result;
				curl_multi_remove_handle( curlMulti, easy );
				if( ( easyResult != CURLE_URL_MALFORMAT ) && ( easyResult != CURLE_OK ) )
				{
					// serious error occured.
					result = false;
					break;
				}
				// prepare and resurrect easy handle for next transfer:
				if( file != files.end() )
				{
					std::string url = host + path + files.GetRelativePath( (*file)->GetPathname() ).c_str();
					__int64 fileSize;
					bool fileSizeAcquired = (*file)->GetSize( fileSize );
					if( fileSizeAcquired )
						curl_easy_setopt( easy, CURLOPT_INFILESIZE_LARGE  , fileSize                     );
					curl_easy_setopt( easy, CURLOPT_URL                   , url.c_str()                  );
					curl_easy_setopt( easy, CURLOPT_READFUNCTION          , CurlReadCallback             );
					curl_easy_setopt( easy, CURLOPT_READDATA              , *file                        );
					curl_multi_add_handle( curlMulti, easy );
					file++;
					running_handles++;
				}
			}
			curlMsg = curl_multi_info_read( curlMulti, &num_messages );
		}

		if( !running_handles )
			break;

		// get curl timeout setting:
		//
		long timeout;	// millisecs
		curl_multi_timeout( curlMulti, &timeout );
		if( timeout < 0 )
			timeout = DEFAULT_TIMEOUT_MILLISECS;
		if( timeout > MAX_TIMEOUT_MILLISECS )
			timeout = MAX_TIMEOUT_MILLISECS;

		// check socket action:
		//
		int fdMax;
		fd_set fdRead, fdWrite, fdException;
		FD_ZERO( &fdRead );
		FD_ZERO( &fdWrite );
		FD_ZERO( &fdException );
		curl_multi_fdset( curlMulti, &fdRead, &fdWrite, &fdException, &fdMax );
		if( fdMax != -1 )
		{
			// socket stuff to do:
			ldiv_t t = ldiv( timeout, 1000L );
			timeval tv;
			tv.tv_sec = t.quot;
			tv.tv_usec = t.rem * 1000;
			int select_result = select( fdMax + 1, &fdRead, &fdWrite, &fdException, &tv );
			if( select_result == SOCKET_ERROR )
			{
				result = false;
				break;	// error
			}
		}
		else
		{
			// Nothing to do on the sockets. 
			// Be nice to other threads and end our time slice here.
			::Sleep( 0 );
		}


	}

	// clean-up libcurl easy handles:
	for( int i = 0; i < numChannels; i++ )
	{
		curl_multi_remove_handle( curlMulti, curlEasy[ i ] );
		curl_easy_cleanup( curlEasy[ i ] );
	}
	delete[] curlEasy;

	// clean-up libcurl multi handle:
	curl_multi_cleanup( curlMulti );

	// clean-up share handle:
	curl_share_cleanup( curlShare );
	CloseHandle( g_hMutex );

	return( result );
}

int main(int argc, char* argv[])
{
	curl_global_init( CURL_GLOBAL_ALL );

	OsFileVector files;
	if( files.PopulateFromRootDirectory( ".\\testdata\\" ) )
		Upload( 2, false, files, "ftp://foo.bar.com", "/testdata", "user", "pass" );

	curl_global_cleanup();

	return 0;
}
