#include "OsFile.h"

OsFile::OsFile( const std::string& strPathname /* = "" */ ) :
	m_strPathname( strPathname ), m_fd( (FILE*)0 )
{
}

OsFile::~OsFile()
{
	Close();
}

bool OsFile::SetPathname( const std::string& strPathname )
{
	if( IsOpen() )
		return( false );

	m_strPathname = strPathname;

	return( true );
}

bool OsFile::Open( const std::string& strMode )
{
	// Assure we are closed:
	if( !Close() )
		return( false );
	
	// Try to open:
	m_fd = fopen( m_strPathname.c_str(), strMode.c_str() );
	
	return( IsOpen() );
}

bool OsFile::GetSize( __int64& size )
{
	FILE* fd = m_fd;

	if( fd == (FILE*)0 )
	{
		fd = fopen( m_strPathname.c_str(), "rb" );
		if( fd == (FILE*)0 )
			return( false );
	}
	
	__int64 currentPos = _ftelli64( fd );
	_fseeki64( fd, 0, SEEK_END );
	size = _ftelli64( fd );
	_fseeki64( fd, currentPos, SEEK_SET );

	if( fd != m_fd )
		fclose( fd );
	
	return( size != -1L );
}

bool OsFile::Close()
{
	if( m_fd ) 
	{
		if( fclose( m_fd ) != 0 )
			return( false );
		
		m_fd = (FILE*)0;

	}
	return( true );
}


//////////////////////////////////////////// OsFileVector


std::string::value_type OsFileVector::cm_dirSep    = (std::string::value_type)'\\';
std::string::value_type OsFileVector::cm_dirSepUrl = (std::string::value_type)'/';

void OsFileVector::ClearAll()
{
	for( iterator iter = begin(); iter != end(); iter++ )
		if( *iter )
			delete *iter;

	clear();
}

OsFileVector::OsFileVector() :
	m_strRootDirectory( "" )
{
}

OsFileVector::~OsFileVector()
{
	ClearAll();
}

bool OsFileVector::PopulateFromDirectory( const std::string& strDirectory, int level )
{
	// All relative to m_strRootDirectory !!

	assert( level < 50 );	// recursion-gone-wild check

	WIN32_FIND_DATA fd;
	std::vector<std::string> dirs;

	std::string strFindMask = m_strRootDirectory + strDirectory + "*";
	HANDLE hFind = FindFirstFile( strFindMask.c_str(), &fd );
	if( hFind == INVALID_HANDLE_VALUE )
		return( false );

	do {
		std::string strName = fd.cFileName;
		
		// ignore special directory entries:
		if( ( strName == "." ) || ( strName == ".." ) )
			continue;
		
		// make path:
		std::string strPath = strDirectory + strName;

		if( fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
		{
			// It's a directory. Keep name for later recursion.
			dirs.push_back( strPath + cm_dirSep );
		}
		else
		{
			// It's a file. Store path:
			push_back( new OsFile( m_strRootDirectory + strPath ) );
		}
	} while( FindNextFile( hFind, &fd ) );

	FindClose( hFind );

	// now recurs into subdirs:
	for( std::vector<std::string>::const_iterator iter = dirs.begin(); iter != dirs.end(); iter++ )
		if( !PopulateFromDirectory( *iter, level + 1 ) )
			return( false );

	dirs.clear();

	return( true );
}

bool OsFileVector::PopulateFromRootDirectory( const std::string& strRootDirectory )
{
	if( strRootDirectory.empty() )
		return( false );

	// make path absolute:
	std::string::value_type* szAbsPath = _fullpath( NULL, strRootDirectory.c_str(), 0 );
	if( !szAbsPath )
		return( false );

	m_strRootDirectory = szAbsPath;
	free( szAbsPath );

	if( m_strRootDirectory[ m_strRootDirectory.length() - 1 ] != cm_dirSep )
		m_strRootDirectory += cm_dirSep;

	ClearAll();

	return( PopulateFromDirectory( "", 0 ) );
}

std::string OsFileVector::GetRelativePath( const OsFile& osFile, bool convertSeparators /* = true */ ) const
{
	std::string relPath;
	if( osFile.GetPathname().find( m_strRootDirectory ) == 0 )
		relPath = osFile.GetPathname().substr( m_strRootDirectory.length() );
	else
		relPath = osFile.GetPathname();	// ??? should never happen
	if( convertSeparators ) {
		for( std::string::iterator iter = relPath.begin(); iter != relPath.end(); iter++ )
			if( (*iter) == cm_dirSep )
				(*iter) = cm_dirSepUrl;
	}
	return( relPath );
}

bool OsFileVector::GetSize( __int64& totalSize )
{
	totalSize = 0;

	for( iterator iter = begin(); iter != end(); iter++ )
	{
		__int64 fileSize;
		if( !(*iter)->GetSize( fileSize ) )
			return( false );

		totalSize += fileSize;
	}

	return( true );
}
